@@ -2,11 +2,14 @@ import abc
 import asyncio
 import collections
 import dataclasses
+import heapq
+import itertools
+import logging
 import typing as t
 
 from n_edit import model
 
-_T = t.TypeVar("_T")
+log = logging.getLogger(__name__)
 
 
 class NodeExecutor(abc.ABC):
@@ -27,10 +30,11 @@ class NodeNotRunnable(Exception):
     pass
 
 
-@dataclasses.dataclass
+@dataclasses.dataclass(order=True)
 class _TaskDefinition:
-    node: model.Node
-    inputs: dict[str, t.Any]
+    priority: int
+    node: model.Node = dataclasses.field(compare=False)
+    inputs: dict[str, t.Any] = dataclasses.field(compare=False)
 
 
 @dataclasses.dataclass
@@ -38,8 +42,17 @@ class _EdgeResult:
     use: t.Callable[[], t.Any]
 
 
+def _const_callable(value: t.Any) -> t.Callable[[], t.Any]:
+    return lambda: value
+
+
 class Run:
-    def __init__(self, graph: model.Graph, node_executor: NodeExecutor) -> None:
+    def __init__(
+        self,
+        graph: model.Graph,
+        node_executor: NodeExecutor,
+        pending_tasks_limit: t.Optional[int] = 10,
+    ) -> None:
         self.graph = graph
         self.node_executor = node_executor
         self.edge_queues: dict[int, collections.deque[t.Any]] = {
@@ -49,15 +62,26 @@ class Run:
         self._static_result_cache: dict[int, t.Any] = {}
         self._executed_nodes: set[int] = set()
         self.pending_tasks: set[asyncio.Task] = set()
+        self.pending_tasks_limit = pending_tasks_limit
 
-        self._tasks_to_create: collections.deque[_TaskDefinition] = collections.deque(
+        self._tasks_to_create: list[_TaskDefinition] = [
             self._create_task_definition(source_node)
             for source_node in self.graph.source_nodes()
-        )
+        ]
+        heapq.heapify(self._tasks_to_create)
+
+        self._task_sequence = itertools.count()
 
     async def run(self) -> None:
         await self._create_pending_tasks()
         while self.pending_tasks:
+            log.debug(
+                "pending tasks: %s, tasks to create: %s, static results: %s, task node id map: %s",
+                len(self.pending_tasks),
+                len(self._tasks_to_create),
+                len(self._static_result_cache),
+                len(self.task_node_id_map),
+            )
             done, self.pending_tasks = await asyncio.wait(
                 self.pending_tasks, return_when=asyncio.FIRST_COMPLETED
             )
@@ -65,9 +89,14 @@ class Run:
                 self.handle_finished_task(task)
             await self._create_pending_tasks()
 
+    def _is_task_limit_reached(self) -> bool:
+        if self.pending_tasks_limit is None:
+            return False
+        return len(self.pending_tasks) >= self.pending_tasks_limit
+
     async def _create_pending_tasks(self) -> None:
-        while self._tasks_to_create:
-            await self._create_task(self._tasks_to_create.popleft())
+        while self._tasks_to_create and not self._is_task_limit_reached():
+            await self._create_task(heapq.heappop(self._tasks_to_create))
 
     def _append_results(self, node_id: int, results: t.Any) -> None:
         node = self.graph.nodes[node_id]
@@ -90,6 +119,7 @@ class Run:
 
     def handle_finished_task(self, task: asyncio.Task) -> None:
         node_id = self.task_node_id_map[task]
+        del self.task_node_id_map[task]
         self._append_results(node_id, task.result())
         for node in self.graph.downstream_nodes(node_id):
             while True:
@@ -97,15 +127,16 @@ class Run:
                     task_definition = self._create_task_definition(node)
                 except NodeNotRunnable:
                     break
-                self._tasks_to_create.append(task_definition)
+                heapq.heappush(self._tasks_to_create, task_definition)
                 self._executed_nodes.add(node.id_)
 
     def _create_task_definition(self, node: model.Node) -> _TaskDefinition:
         inputs = self._collect_all_input_values(node.id_)
-        return _TaskDefinition(node, inputs)
+        priority = self.graph.distance_from_end(node.id_)
+        return _TaskDefinition(priority, node, inputs)
 
     async def _create_task(self, task_definition: _TaskDefinition) -> asyncio.Task:
-        name = f"{task_definition.node.type_name}.{task_definition.node.id_}({task_definition.inputs})"
+        name = f"{task_definition.node.type_name()}.{task_definition.node.id_}({next(self._task_sequence)})"
         task = asyncio.create_task(
             self.node_executor.execute_node(
                 task_definition.node, task_definition.inputs
@@ -134,12 +165,12 @@ class Run:
                 static_result = self._static_result_cache[edge.id_]
             except KeyError:
                 continue
-            result[slot_name] = _EdgeResult(lambda: static_result)
+            result[slot_name] = _EdgeResult(_const_callable(static_result))
         return result
 
     def _collect_defaults(self, node_id: int) -> dict[str, _EdgeResult]:
         return {
-            name: _EdgeResult(lambda: default)
+            name: _EdgeResult(_const_callable(default))
             for name, default in self.graph.get_slot_defaults(node_id).items()
         }
 
