@@ -187,6 +187,14 @@ class Graph(serializable.Serializable):
         node = self.nodes[node_id]
         return node.is_static(self.upstream_nodes(node_id), self)
 
+    def distance_from_end(self, node_id: int) -> int:
+        child_node_ids = [
+            edge.end.owner.id_ for edge in self.get_outgoing_edges(node_id)
+        ]
+        if not child_node_ids:
+            return 0
+        return min(self.distance_from_end(id_) for id_ in child_node_ids) + 1
+
 
 _T = t.TypeVar("_T")
 
@@ -243,7 +251,9 @@ class Node(t.Generic[_T]):
     def is_static(self, inputs: dict[str, Node], graph: Graph) -> bool:
         if not inputs:
             return not self.has_iterable_output
-        return all(graph.is_node_static(node.id_) for node in inputs.values())
+        return not self.has_iterable_output and all(
+            graph.is_node_static(node.id_) for node in inputs.values()
+        )
 
     @classmethod
     def type_name(cls) -> str:
